#include <string.h>
#include <keyvalpb.h>
#include <stdio.h>
#include <stdlib.h>
#include <utils.h>
#include <malloc.h>
#include "keyval_cmd.h"


#define KVPB_CHECK_VALIDITY

/**
 * kvpb_store_init - Open, check and initialize structures to access key-value parameter block
 * @store: Pointer to the KVPB access information/state structure
 * @base: KVPB storage base address of the first mapped region.
 * @size: Size of one region of the KVPB storage.
 * @flags: KVPB storage flags, typical are %KVPB_DESC_DOUBLE and %KVPB_DESC_ALIGN4
 * @mode: Initialization mode: 0 .. initialize and check only, 1 .. try to correct
 *	possible inconsistency or if data cannot be repaired, erase whole storage,
 *	2 .. erase data unconditionally
 *
 * Return Value: Value returned by kvpb_check() or -1 if the block parameters
 *	are incorrect
 * File: keyval_cmd.c
 */
int kvpb_store_init(kvpb_block_t *store, void *base, int size, int flags, int mode)
{
  int res;
  if(!base)
    return -1;
  store->base=base;
  if(!store->base)
    return -1;
  store->size=size;
  store->flags=flags;
  if(mode & 2){
    kvpb_block_erase(store, store->base,store->size);
    if(store->flags & KVPB_DESC_DOUBLE)
      kvpb_block_erase(store, (uint8_t*)store->base+store->size,store->size);
  }
  res=kvpb_check(store,mode);
  /*printf("kvpb_init_store: kvpb_check returned %d\n", res);*/

  return res;
}


/**
 * kvpb_store_list - List KVPB storage content
 * @store: Pointer to the KVPB access information/state structure
 * @mode: Unused for now
 *
 * Return Value: Negative value inform about unsuccessful operation.
 * File: keyval_cmd.c
 */
int kvpb_store_list(kvpb_block_t *store, int mode)
{
  kvpb_key_t *key;
  int i;
  uint8_t *p;
 #ifdef KVPB_CHECK_VALIDITY
  uint8_t *base=(uint8_t*)store->base;
  uint8_t *end=(uint8_t*)store->psum1;
  if((store->flags & KVPB_DESC_USE2ND)&&(store->flags&KVPB_DESC_DOUBLE)){
    base+=store->size;
    end=(uint8_t*)store->psum2;
  }
 #endif /*KVPB_CHECK_VALIDITY*/
  
  kvpb_for_each(store, key, 0) {
    printf("key 0x%08x size %4d :", (int)key->keyid, (int)key->size);
    
   #ifdef KVPB_CHECK_VALIDITY
    if((((uint8_t*)key+key->size)>end)||
       ((uint8_t*)key<base)||((uint8_t*)key>=end)) {
      printf("\nERROR keyval corrupted, setting read only\n");
      store->flags|=KVPB_DESC_RO;
      return -1;
    }
   #endif /*KVPB_CHECK_VALIDITY*/
    p=(uint8_t *)kvpb_key2data(key);
    for(i=0;i<key->size;i++){
      printf("%02x ",*(p++));
    }
    printf("\n");
  }
  return 0;
}


/**
 * kvpb_get_key_dynamic - Allocate buffer and get value for given key ID
 * @store: Pointer to the KVPB access information/state structure
 * @keyid: Ordinal value representing key ID
 * @psize: Pointer to the place where retrieved value size should be stored
 * @pbuf: Pointer to the place where pointer to the allocated value buffer should be stored
 *
 * Return Value: Number of retrieved value bytes if operation is successful
 *	or -1 if there is no such key ID or operation fails for other reason.
 *	If space for buffer cannot be allocated -3 is returned.
 * File: keyval_cmd.c
 */
int kvpb_get_key_dynamic(kvpb_block_t *store, kvpb_keyid_t keyid, kvpb_size_t *psize, void **pbuf)
{
  kvpb_size_t size=0;
  void *buf=NULL;
  int  res=0;

  *psize=0;
  *pbuf=NULL;

  do{
    if(buf) free(buf);
    if(size){
      buf=malloc(size);
      if(!buf)
	return -3;
    }

    res=kvpb_get_key(store, keyid, size, buf);
    if(res<0){
      if(buf)
        free(buf);
      return res;
    }
    if(size>=res)
      break;
    size=res;
  }while(1);

  if(psize)
    *psize=res;
  *pbuf=buf;
  return res;
}


/**
 * kvpb_stress_test - Command line interface function to stress test store
 * @store: Pointer to the KVPB access information/state structure
 * @cycles: Number of cycles to run
 * @min_size: Minimal test block size
 * @max_size: Maximal test block size
 * @min_id: Minimal key ID
 * @max_id: Maximal key ID
 *
 * Return Value: Negative value inform about unsuccessful operation.
 * File: keyval_cmd.c
 */
int kvpb_stress_test(kvpb_block_t *store, int cycles, kvpb_size_t min_size,
                     kvpb_size_t max_size, kvpb_keyid_t min_id, kvpb_keyid_t max_id)
{
  int ccnt = 0;
  char *buff;
  kvpb_keyid_t id;
  kvpb_size_t size;
  int ret;

  if((min_size > max_size) || (min_id > max_id))
    return -1;

  buff = malloc(max_size);

  while(++ccnt <= cycles) {
    size = rand() % (max_size - min_size + 1) + min_size;
    id = rand() % (max_id - min_id + 1) + min_id;
    snprintf(buff, size, "id %d, sz %d c %d ABCDEFGHIJKLMNOPQRSTUVWXYZ",
             (int)id, (int)size, ccnt);
    ret = kvpb_set_key(store, id, size, buff);
    if(ret < 0) {
      fprintf(stderr,"cycle #%d kvpb_set_key failed for id %d and size %d\n",
              ccnt, (int)id, (int)size);
    }
  }
  free(buff);

  return 0;
}

/**
 * kvpb_get_cmd - Command line interface function to retrieve value for given key
 * @store: Pointer to the KVPB access information/state structure
 * @line: Command line parameters
 *
 * Return Value: Negative value inform about unsuccessful operation.
 * File: keyval_cmd.c
 */
int kvpb_get_cmd(kvpb_block_t *store, char *line)
{
  char *p;
  uint8_t *pu;
  int  res;
  long val;
  kvpb_keyid_t keyid;
  kvpb_size_t arr_len=0;
  void *arr_data=NULL;

  p=line;

  si_skspace(&p);
  if(!*p) return 0;

  if(si_long(&p, &val, 0)<=0){
    printf("kvpb_get_cmd: bad keyid format\n");
    return -2;
  }
  keyid=val;

  if(keyid&KVPB_KEYID_DUPLIC){
    int i;
    res=0;
    kvpb_key_t *key;
    kvpb_each_from(store, keyid, 0, key) {
      printf("key 0x%08x size %4d :", (int)key->keyid, (int)key->size);
      pu=(uint8_t *)kvpb_key2data(key);
      for(i=0;i<key->size;i++){
	printf("%02x ",*(pu++));
      }
      printf("\n");
    }
  }
  else {
    int i;
    res=kvpb_get_key_dynamic(store, keyid, &arr_len, &arr_data);
    if(res<0){
      printf("ERROR: kvpb_get_key returned %d\n",res);
      return res;
    }

    printf("key 0x%08x size %4d :", (int)keyid, (int)arr_len);

    pu=(uint8_t *)arr_data;
    for(i=0;i<arr_len;i++){
      printf("%02x ",*(pu++));
    }
    printf("\n");
    if(arr_data)
      free(arr_data);
  }

  return res;
}

/**
 * kvpb_get_cmd - Command line interface function to set value for given key
 * @store: Pointer to the KVPB access information/state structure
 * @line: Command line parameters
 *
 * Return Value: Negative value inform about unsuccessful operation.
 * File: keyval_cmd.c
 */
int kvpb_set_cmd(kvpb_block_t *store, char *line)
{
  char *p;
  int  res;
  long val;
  kvpb_keyid_t keyid;
  int arr_len=0;
  void *arr_data=NULL;

  p=line;
    
  si_skspace(&p);
  if(!*p) return 0;
  
  if(si_long(&p, &val, 0)<=0){
    printf("kvpb_set_cmd: bad keyid format\n");
    return -2;
  }
  keyid=val;

  if(si_add_to_arr(&p, &arr_data, &arr_len, 0, 1, NULL)<0){
    printf("kvpb_set_cmd: bad data format\n");
    if(arr_data)
      free(arr_data);
    return -2;
  }

  res=kvpb_set_key(store, keyid, arr_len, arr_data?arr_data:&arr_len);
  if(arr_data)
    free(arr_data);
  if(res<0){
    printf("ERROR: kvpb_set_key returned %d\n",res);
  }
  return res;
}

/**
 * kvpb_del_cmd - Command line interface function to delete given key and values
 * @store: Pointer to the KVPB access information/state structure
 * @line: Command line parameters
 *
 * Return Value: Negative value inform about unsuccessful operation.
 * File: keyval_cmd.c
 */
int kvpb_del_cmd(kvpb_block_t *store, char *line)
{
  char *p;
  int  res;
  long val;
  kvpb_keyid_t keyid;

  p=line;

  si_skspace(&p);
  if(!*p) return 0;

  if(si_long(&p, &val, 0)<=0){
    printf("kvpb_del_cmd: bad keyid format\n");
    return -2;
  }
  keyid=val;

  res=kvpb_set_key(store, keyid, 0, NULL);
  if(res<0){
    printf("ERROR: kvpb_set_key returned %d\n",res);
  }

  return res;
}
