/*******************************************************************
  Key Value Persistent Storage

  keyval_dilist.c	- names through IDs storage and retrieval

  (C) Copyright 2003-2005 by Pavel Pisa - Originator

  The uLan utilities library can be used, copied and modified under
  next licenses
    - GPL - GNU General Public License
    - LGPL - GNU Lesser General Public License
    - MPL - Mozilla Public License
    - and other licenses added by project originators
  Code can be modified and re-distributed under any combination
  of the above listed licenses. If contributor does not agree with
  some of the licenses, he/she can delete appropriate line.
  Warning, if you delete all lines, you are not allowed to
  distribute source code and/or binaries utilizing code.
  
  See files COPYING and README for details.

 *******************************************************************/

#include <inttypes.h>
#include <string.h>
#include <keyvalpb.h>
#include <stdio.h>
#include <malloc.h>
#include <limits.h>
#include <keyval_dilist.h>
#include <keyval_cmd.h>
#include <keyval_id.h>

#include <ul_list.h>
#include <ul_gavl.h>
#include <ul_gavlcust.h>
#include <ul_uniqid.h>

UL_LIST_CUST_DEC(kvpb_di_list_withid, kvpb_di_list_t, kvpb_di_item_t,
                withid_list, withid_node)

UL_LIST_CUST_DEC(kvpb_di_list_noid, kvpb_di_list_t, kvpb_di_item_t,
                noid_list, withid_node)

static inline int
kvpb_di_id_cmp_fnc( const kvpb_di_id_t *a, const kvpb_di_id_t *b)
{
        if (*a>*b) return 1;
        if (*a<*b) return -1;
        return 0;
}

GAVL_CUST_NODE_INT_DEC(kvpb_di_list_id, kvpb_di_list_t, kvpb_di_item_t, kvpb_di_id_t,
                id_root, id_node, id, kvpb_di_id_cmp_fnc)
GAVL_CUST_NODE_INT_IMP(kvpb_di_list_id, kvpb_di_list_t, kvpb_di_item_t, kvpb_di_id_t,
                id_root, id_node, id, kvpb_di_id_cmp_fnc)


static inline int
kvpb_di_name_cmp_fnc( kvpb_di_name_t *a, kvpb_di_name_t *b)
{
	int res;

        res=memcmp(a->str, b->str, (a->len < b->len)? a->len: b->len);
	if(res)
	  return res;
        if (a->len>b->len) return 1;
        if (a->len<b->len) return -1;
        return 0;
}

GAVL_CUST_NODE_INT_DEC(kvpb_di_list_name, kvpb_di_list_t, kvpb_di_item_t, kvpb_di_name_t,
                name_root, name_node, name, kvpb_di_name_cmp_fnc)
GAVL_CUST_NODE_INT_IMP(kvpb_di_list_name, kvpb_di_list_t, kvpb_di_item_t, kvpb_di_name_t,
                name_root, name_node, name, kvpb_di_name_cmp_fnc)

kvpb_di_item_t *kvpb_di_list_new_item(kvpb_di_name_t *name)
{
  kvpb_di_item_t *diitem;
  int name_len;
  char *p;
  
  diitem=malloc(sizeof(kvpb_di_item_t));
  if(!diitem)
    return NULL;
  memset(diitem,0,sizeof(kvpb_di_item_t));
  if(name->str){
    name_len=name->len;
    if(name_len<=0)
      name_len=strlen(name->str);
    p=malloc(name_len+1);
    if(!p){
      free(diitem);
      return NULL;
    }
    diitem->name.len=name_len;
    memcpy(p,name->str,name_len);
    p[name_len]=0;
    diitem->name.str=p;
    diitem->flags|=KVPB_DI_NAME;
  }else{
    diitem->flags|=KVPB_DI_FREE;
  }
  return diitem;
}

void kvpb_di_list_free_item(kvpb_di_item_t *diitem)
{
  if(diitem->name.str)
    free((char*)diitem->name.str);
  free(diitem);
}

int kvpb_di_list_setid_internal(struct kvpb_di_list *dilist,
	kvpb_di_item_t *diitem, kvpb_di_id_t id)
{
  diitem->id=id;
  if(ul_uniqid_pool_reserve(&dilist->idpool, id, id)<=0){
    kvpb_di_item_t *diitem_sameid;
    diitem_sameid=kvpb_di_list_id_find(dilist,&id);
    if(diitem_sameid){
      if(!(diitem_sameid->flags&KVPB_DI_FORCED)){
	kvpb_di_list_id_delete(dilist,diitem_sameid);
	diitem_sameid->id=0;
	kvpb_di_list_withid_delete(dilist,diitem_sameid);
	kvpb_di_list_noid_insert(dilist,diitem_sameid);
      }
    }
  }
  if(kvpb_di_list_id_insert(dilist,diitem)<0)
    diitem->id=0;

  if(diitem->id){
    kvpb_di_list_withid_insert(dilist,diitem);
  }else{
    kvpb_di_list_noid_insert(dilist,diitem);
  }

  return 0;
}

kvpb_di_item_t *kvpb_di_list_add_internal(struct kvpb_di_list *dilist,
	kvpb_di_name_t *name, int min_idx, int max_idx, int flags, kvpb_di_id_t id)
{
  kvpb_di_item_t *diitem;
  int addonreq=0;

  diitem=kvpb_di_list_name_find(dilist,name);
  if(diitem){
    if(flags&KVPB_DI_RDWR){
      if(!(diitem->flags&KVPB_DI_REQ))
        addonreq=1;
      diitem->flags|=KVPB_DI_RDWR;
      diitem->flags&=~KVPB_DI_DEL;
      if(diitem->min_idx > min_idx)
        diitem->min_idx=min_idx;
      if(diitem->max_idx < max_idx)
        diitem->max_idx=max_idx;
    }
    if(flags&KVPB_DI_DEL){
      if(!(diitem->flags&KVPB_DI_REQ))
        addonreq=1;
      diitem->flags|=KVPB_DI_DEL;
    }

    if(id && (diitem->id != id) && !(diitem->flags&KVPB_DI_FORCED)) {
      if(diitem->id){
        /* FIXME for duplicate name there is memory leak */
        ul_uniqid_pool_free_one(&dilist->idpool, id);
        kvpb_di_list_id_delete(dilist,diitem);
        kvpb_di_list_withid_delete(dilist,diitem);
      }else{
        kvpb_di_list_noid_delete(dilist,diitem);
      }

      kvpb_di_list_setid_internal(dilist,diitem,id);
    }
    diitem->flags|=flags&KVPB_DI_FORCED;
  } else {
    diitem=kvpb_di_list_new_item(name);
    if(!diitem)
      return NULL;
    diitem->flags=flags;
    diitem->id=0;
    diitem->min_idx=min_idx;
    diitem->max_idx=max_idx;
    if(kvpb_di_list_name_insert(dilist,diitem)<0){
      /* Fatal error, cannot happen*/
      return NULL;
    }

    if(id) {
      kvpb_di_list_setid_internal(dilist,diitem,id);
    }else{
      kvpb_di_list_noid_insert(dilist,diitem);
    }
    
    if(flags&KVPB_DI_REQ)
      addonreq=1;
  }

  if(addonreq){
    kvpb_di_list_req_insert(dilist,diitem);
  }

  return diitem;
}

/**
 * kvpb_di_list_add - Add new named item into conversion list
 * @dilist:	pointer to the conversion list head
 * @name_str:	ASCII name to add onto list
 * @name_len:	length of name or -1 if length should be computed for
 *		  zero terminated string
 * @min_idx:	minimal index to store for dinfo load/store purposes
 * @max_idx:	maximal index to store
 * @flags:	requested operation flags: %KVPB_DI_RDWR .. default operation;
 *		  %KVPB_DI_DEL .. to delete name from stored conversion table
 *
 * This function returns negative value if some operation fails.
 */
int kvpb_di_list_add(struct kvpb_di_list *dilist,
	char *name_str, int name_len, int min_idx, int max_idx, int flags)
{
  kvpb_di_item_t *diitem;
  kvpb_di_name_t name;

  name.str=name_str;
  if(name_len<=0)
    name_len=strlen(name_str);
  name.len=name_len;
  if(name.len>255)
    return -1;

  if(!flags)
    flags|=KVPB_DI_RDWR;

  flags&=KVPB_DI_REQ;
  diitem=kvpb_di_list_add_internal(dilist, &name, min_idx, max_idx, flags, 0);
  return diitem?0:-1;
}

/**
 * kvpb_di_list_enforce_id - Enforce predefined ID for provided name
 * @dilist:	pointer to the conversion list head
 * @name_str:	ASCII name to add onto list
 * @name_len:	length of name or -1 if length should be computed for
 *		  zero terminated string
 * @diid:	forced ID value
 *
 * This function returns negative value if some operation fails.
 */
int kvpb_di_list_enforce_id(struct kvpb_di_list *dilist, char *name_str, int name_len, kvpb_di_id_t diid)
{
  kvpb_di_item_t *diitem;
  kvpb_di_name_t name;

  if(!diid)
    return -1;

  name.str=name_str;
  if(name_len<=0)
    name_len=strlen(name_str);
  name.len=name_len;
  if(name.len>255)
    return -1;

  diitem=kvpb_di_list_add_internal(dilist, &name, INT_MAX, 0, 0, diid);
  return diitem?0:-1;
}

/**
 * kvpb_di_list_name2id - Return unique ID assigned to the name
 * @dilist:	pointer to the conversion list head
 * @name_str:	ASCII name to add onto list
 * @name_len:	length of name or -1 if length should be computed for
 *		  zero terminated string
 * @pdiid:	pointer to storage for found ID
 *
 * Function can be used to obtain unique ID value for name in the question.
 * if no such name is found in the conversion list, negative value is returned.
 * The function should be called, after unique IDs are consolidated
 * with already stored ones by call to @kvpb_di_list_load_ids and yet unassigned items
 * got new unique IDs by call to @kvpb_di_list_generate_ids
 */
int kvpb_di_list_name2id(const struct kvpb_di_list *dilist, char *name_str, int name_len, kvpb_di_id_t *pdiid)
{
  kvpb_di_item_t *diitem;
  kvpb_di_name_t name;

  *pdiid=0;

  name.str=name_str;
  if(name_len<=0)
    name_len=strlen(name_str);
  name.len=name_len;
  if(name.len>255)
    return -1;
  
  diitem=kvpb_di_list_name_find(dilist, &name);
  if(!diitem)
    return -1;

  *pdiid=diitem->id;

  return 1;
}

/**
 * kvpb_di_list_new - Create new list for preparation of name to IDS conversion
 *
 * This function returns new empty initialized names to IDs conversion list
 */
struct kvpb_di_list *kvpb_di_list_new(void)
{
  kvpb_di_list_t *dilist;
  dilist=malloc(sizeof(kvpb_di_item_t));
  if(!dilist)
    return NULL;
  if(kvpb_di_list_init(dilist)<0){
    free(dilist);
    return NULL;
  }
  return dilist;
}

/**
 * kvpb_di_list_init - Initializes list for preparation of name to IDS conversion
 * @dilist: pointer to conversion list head
 *
 * This returns negative value if initialization fails
 */
int kvpb_di_list_init(kvpb_di_list_t *dilist)
{
  kvpb_di_list_req_init_head(dilist);
  kvpb_di_list_withid_init_head(dilist);
  kvpb_di_list_noid_init_head(dilist);
  kvpb_di_list_id_init_root_field(dilist);
  kvpb_di_list_name_init_root_field(dilist);
  dilist->store=NULL;
  dilist->update_names=0;

  return ul_uniqid_pool_init(&dilist->idpool, KVPB_DI_MINDIID, KVPB_DI_MAXDIID);
}


/**
 * kvpb_di_list_done - Releases items connected to the list
 * @dilist: pointer to conversion list head
 *
 * This returns negative value if destruction fails
 */
int kvpb_di_list_done(kvpb_di_list_t *dilist)
{
  kvpb_di_item_t *diitem;
  ul_uniqid_pool_done(&dilist->idpool);
  
  ul_list_for_each_cut(kvpb_di_list_withid, dilist, diitem)
    kvpb_di_list_free_item(diitem);
  
  ul_list_for_each_cut(kvpb_di_list_noid, dilist, diitem)
    kvpb_di_list_free_item(diitem);
  
  return 0;
}


/**
 * kvpb_di_list_generate_ids - Generate unique IDs for yet not assigned items
 * @dilist:	pointer to the conversion list head
 *
 * The function goes through all items on @noid_list and allocates
 * unique IDs for them from values ranges available in the @idpool
 *
 * The function returns negative value if there is some problem
 */
int kvpb_di_list_generate_ids(kvpb_di_list_t *dilist)
{
  ul_uniqid_t  poolid;
  kvpb_di_item_t *diitem;

  ul_list_for_each_cut(kvpb_di_list_noid, dilist, diitem){
    if(ul_uniqid_pool_alloc_one(&dilist->idpool, &poolid)<0){
      kvpb_di_list_noid_insert(dilist, diitem);
      return -1;
    }
    dilist->update_names=1;
    diitem->id=poolid;
    kvpb_di_list_withid_insert(dilist, diitem);
    kvpb_di_list_id_insert(dilist, diitem);
    
  }
  
  return 0;
}

/**
 * kvpb_di_list_load_ids - Load data into conversion list from the storage block
 * @dilist:	pointer to the conversion list head
 * @store:	pointer to the keyval storage block description
 * @keyid4ids:	keyval ID of the value/block holding conversion table
 *
 * The function returns negative value if there is some problem
 */
int kvpb_di_list_load_ids(kvpb_di_list_t *dilist, struct kvpb_block *store, int keyid4ids)
{
  int res;
  kvpb_size_t data_len=0;
  void *data=NULL;
  kvpb_di_id_t diid;
  uint8_t *p;
  uint8_t *end;
  kvpb_di_name_t name;
  kvpb_di_item_t *diitem;

  /*read and parse KVPB_KEYID_DINFO_NAMES*/
  res=kvpb_get_key_dynamic(store, keyid4ids, &data_len, &data);
  if(res<0)
    return res;
  
  if(!data)
    return 0;

  p=(uint8_t*)data;
  end=(uint8_t*)data+data_len;

  while(p<end-sizeof(kvpb_di_id_t)-1){
    diid=*(kvpb_di_id_t*)p;
    p+=sizeof(kvpb_di_id_t);
    name.len=*(uint8_t*)p;
    if(p+kvpb_align(name.len+1)>end)
      break;
    name.str=(char*)p+1;
    diitem=kvpb_di_list_add_internal(dilist, &name, INT_MAX, 0, KVPB_DI_FORCED, diid);
    p+=kvpb_align(name.len+1);
  }

  free(data);
  
  return 0;
}

/**
 * kvpb_di_list_save_ids - Save conversion list to the storage block
 * @dilist:	pointer to the conversion list head
 * @store:	pointer to the keyval storage block description
 * @keyid4ids:	keyval ID of the value/block holding conversion table
 *
 * This function stores conversion list into keyval storage.
 * All items of the list should have assigned IDs from the previous
 * calls to @kvpb_di_list_load_ids, @kvpb_di_list_generate_ids or
 * @kvpb_di_list_enforce_id.
 *
 * The function returns negative value if there is some problem
 */
int kvpb_di_list_save_ids(kvpb_di_list_t *dilist, struct kvpb_block *store, int keyid4ids)
{
  int res;
  kvpb_size_t data_len;
  void *data=NULL;
  uint8_t *p;
  uint8_t *end;
  kvpb_di_item_t *diitem;

  data_len=0;
  gavl_cust_for_each(kvpb_di_list_name, dilist, diitem){
    if(diitem->flags&KVPB_DI_DEL) continue;
    data_len+=sizeof(kvpb_di_id_t)+kvpb_align(diitem->name.len+1);
  }

  if(data_len){
    data=malloc(data_len);
    if(!data)
      return -1;

    p=(uint8_t*)data;
    end=(uint8_t*)data+data_len;

    gavl_cust_for_each(kvpb_di_list_name, dilist, diitem){
      if(diitem->flags&KVPB_DI_DEL) continue;
      *(kvpb_di_id_t*)p=diitem->id;
      p+=sizeof(kvpb_di_id_t);
      *(uint8_t*)p=diitem->name.len;
      memcpy(p+1,diitem->name.str,diitem->name.len);

      p+=kvpb_align(diitem->name.len+1);
    }

    if(p!=end){
      free(data);
      return -1;
    }
  }

  res=kvpb_set_key(store, keyid4ids, data_len, data);
  if(data)
    free(data);
  return res;
}

/**
 * kvpb_di_list_for_each - Call provided function for each list element.
 * @dilist: pointer to conversion list head
 * @context: user context pointer
 * @callfnc: function to call
 */
int kvpb_di_list_for_each(const struct kvpb_di_list *dilist, void *context,
                           int (*callfnc)(void *context, kvpb_di_item_t *diitem))
{
  kvpb_di_item_t *diitem;
  int res=0;

  gavl_cust_for_each(kvpb_di_list_name, dilist, diitem){
    res=(*callfnc)(context, diitem);
    if(res<0)
      break;
  }
  return res;
}

/**
 * kvpb_di_list_print - Prints elements of conversion list
 * @dilist: pointer to conversion list head
 */
int kvpb_di_list_print(kvpb_di_list_t *dilist)
{
  kvpb_di_item_t *diitem;

  gavl_cust_for_each(kvpb_di_list_name, dilist, diitem){
    printf("0x%04x \"%s\"\n",diitem->id, diitem->name.str);
  }
  return 0;
}
