/*******************************************************************
  Key Value Persistent Storage

  keyval_dilist.h	- names through IDs storage and retrieval

  (C) Copyright 2003-2005 by Pavel Pisa - Originator

  The uLan utilities library can be used, copied and modified under
  next licenses
    - GPL - GNU General Public License
    - LGPL - GNU Lesser General Public License
    - MPL - Mozilla Public License
    - and other licenses added by project originators
  Code can be modified and re-distributed under any combination
  of the above listed licenses. If contributor does not agree with
  some of the licenses, he/she can delete appropriate line.
  Warning, if you delete all lines, you are not allowed to
  distribute source code and/or binaries utilizing code.
  
  See files COPYING and README for details.

 *******************************************************************/

#ifndef _KEYVAL_DILIST_H
#define _KEYVAL_DILIST_H

#include <ul_list.h>
#include <ul_gavl.h>
#include <ul_uniqid.h>

typedef int kvpb_di_id_t;

#define KVPB_DI_MINDIID 1
#define KVPB_DI_MAXDIID 0x7fff
#define KVPB_DI_ARRAYFL 0x8000

#define KVPB_DI_FREE 1
#define KVPB_DI_NAME 2
#define KVPB_DI_RDWR 4
#define KVPB_DI_DEL  8
#define KVPB_DI_REQ  (KVPB_DI_RDWR|KVPB_DI_DEL)
#define KVPB_DI_FORCED 16
#define KVPB_DI_DIFFER 32

typedef struct kvpb_di_name {
  const char *str;
  int         len;
} kvpb_di_name_t;

/**
 * struct kvpb_di_item - Helper structure for name to unique ID conversion
 * @flags:	item flags: %KVPB_DI_RDWR .. request to read or write item,
 *		  %KVPB_DI_DEL .. request to delete item and ID,
 *		  %KVPB_DI_FORCED .. value of ID is forced and replaces other items
 * @id:		predefined forced or computed unique item ID
 * @name:	ASCII name of item
 * @min_idx:	minimal index to take care of
 * @max_idx:	maximal index to take care of
 * @req_node:	kvpb_di_list_req node
 * @withid_node: node to put item on withid or noid lists
 * @id_node:	node for GAVL tree sorted by IDs
 * @name_node:	node for GAVL tree sorted by names
 */
typedef struct kvpb_di_item {
  int flags;
  kvpb_di_id_t id;
  kvpb_di_name_t name;
  int min_idx;
  int max_idx;

  ul_list_node_t req_node;
  ul_list_node_t withid_node;
  gavl_node_t id_node;
  gavl_node_t name_node;
} kvpb_di_item_t;


/**
 * struct kvpb_di_list - List of item names for their conversion to unique IDs
 * @store:	pointer to data storage block description
 * @update_names: set by @kvpb_di_list_generate_ids when stored name to ID
 *		  conversion list needs update
 * @req_list:	list of requested name to ID items
 * @withid_list: list of named items, which has already assigned or computed unique ID
 * @noid_list:	list of items, which has not been assigned IDs yet
 * @id_root:	GAVL tree of items sorted by IDs
 * @name_root:	GAVL tree of items sorted by their names
 * @idpool:	pool to allocate unique IDs from
 */
typedef struct kvpb_di_list {
  struct kvpb_block *store;
  int update_names;
  
  ul_list_head_t req_list;
  ul_list_head_t withid_list;
  ul_list_head_t noid_list;
  gavl_cust_root_field_t id_root;
  gavl_cust_root_field_t name_root;
  ul_uniqid_pool_t idpool;
} kvpb_di_list_t;


UL_LIST_CUST_DEC(kvpb_di_list_req, kvpb_di_list_t, kvpb_di_item_t,
                req_list, req_node)


struct kvpb_block;

struct kvpb_di_list *kvpb_di_list_new(void);
int kvpb_di_list_init(struct kvpb_di_list *dilist);
int kvpb_di_list_done(struct kvpb_di_list *dilist);

int kvpb_di_list_print(struct kvpb_di_list *dilist);

int kvpb_di_list_add(struct kvpb_di_list *dilist, char *name, int name_len, int min_idx, int max_idx, int flags);
int kvpb_di_list_enforce_id(struct kvpb_di_list *dilist, char *name, int name_len, kvpb_di_id_t diid);
int kvpb_di_list_name2id(const struct kvpb_di_list *dilist, char *name, int name_len, kvpb_di_id_t *diid);

int kvpb_di_list_generate_ids(struct kvpb_di_list *dilist);
int kvpb_di_list_load_ids(struct kvpb_di_list *dilist, struct kvpb_block *store, int keyid4ids);
int kvpb_di_list_save_ids(struct kvpb_di_list *dilist, struct kvpb_block *store, int keyid4ids);

kvpb_di_item_t *kvpb_di_list_id_find(const struct kvpb_di_list *dilist, const kvpb_di_id_t *id);
kvpb_di_item_t *kvpb_di_list_id_find_after(const struct kvpb_di_list *dilist, const kvpb_di_id_t *id);
kvpb_di_item_t *kvpb_di_list_name_find(const struct kvpb_di_list *dilist, const kvpb_di_name_t *name);
kvpb_di_item_t *kvpb_di_list_name_find_after(const struct kvpb_di_list *dilist, const kvpb_di_name_t *name);

int kvpb_di_list_for_each(const struct kvpb_di_list *dilist, void *context,
                           int (*callfnc)(void *context, kvpb_di_item_t *diitem));

#endif /*_KEYVAL_DILIST_H*/
