/*******************************************************************
  Key Value Persistent Storage

  keyval_dinfo.c	- dinfo storage/retrieve 

  (C) Copyright 2003-2005 by Pavel Pisa - Originator

  The uLan utilities library can be used, copied and modified under
  next licenses
    - GPL - GNU General Public License
    - LGPL - GNU Lesser General Public License
    - MPL - Mozilla Public License
    - and other licenses added by project originators
  Code can be modified and re-distributed under any combination
  of the above listed licenses. If contributor does not agree with
  some of the licenses, he/she can delete appropriate line.
  Warning, if you delete all lines, you are not allowed to
  distribute source code and/or binaries utilizing code.
  
  See files COPYING and README for details.

 *******************************************************************/

#include <inttypes.h>
#include <string.h>
#include <keyvalpb.h>
#include <stdio.h>
#include <malloc.h>
#include <limits.h>
#include <keyval_dinfo.h>
#include <keyval_dilist.h>
#include <keyval_cmd.h>
#include <keyval_id.h>
#include <keyval_dbuff.h>
#include <suiut/sui_dinfo.h>
#include <suiut/sui_di_text.h>
#include <suiut/sui_utf8.h>
#include <suiut/namespace.h>

#include <ul_dbuff.h>
#include <ul_log.h>

UL_LOG_CUST(ulogd_keyval_dinfo)

static inline void
kvpb_serialize_internal(void **pp, const void *data, int size)
{
  char *p=(char*)*pp;
  memcpy(p,data,size);
  *pp=p+size;
}

static inline void
kvpb_deserialize_internal(const void **pp, void *data, int size)
{
  char *p=(char*)*pp;
  memcpy(data,p,size);
  *pp=p+size;
}

static int
kvpb_dinfos2dbuff_utf8(ul_dbuff_t *pdbuf, sui_typeid_t diid, sui_dinfo_t *dinfo, int startidx, int count, int options)
{
  int size, fullsize;
  sui_typeid_t tinfo=SUI_TYPE_UTF8;
  int prev_len;
  void *p;
  int i;
  int res=0;
  utf8 *utf8_array[count];

  for(i=0;i<count;i++)
    utf8_array[i]=NULL;

  do{
    fullsize=sizeof(fullsize)+sizeof(tinfo)+sizeof(diid);
    if(startidx){
      diid|=KVPB_DI_ARRAYFL;
      fullsize+=sizeof(startidx);
    }

    for(i=0;i<count;i++){
      if(sui_rd_utf8(dinfo, startidx+i, &utf8_array[i])!=SUI_RET_OK) {
        if(!(options & KVPB_DINFOS_IGNORE_ERR_VAL)) {
	  res = -1;
	  break;
	}
	size=0;
	fullsize+=sizeof(size)+size;
	utf8_array[i]=NULL;
      } else {
        size=sui_utf8_bytesize(utf8_array[i]);
        fullsize+=sizeof(size)+size;
      }
    }

    /* We need to comply with alignment rules */
    fullsize=(fullsize+3)&~3;

    prev_len=pdbuf->len;
    if(ul_dbuff_set_len(pdbuf, pdbuf->len+fullsize)<=0){
      res = -1;
      break;
    }
    p=pdbuf->data + prev_len;

    kvpb_serialize_internal(&p, &fullsize, sizeof(fullsize));
    kvpb_serialize_internal(&p, &tinfo, sizeof(tinfo));
    kvpb_serialize_internal(&p, &diid, sizeof(diid));
    if(diid&KVPB_DI_ARRAYFL)
      kvpb_serialize_internal(&p, &startidx, sizeof(startidx));

    for(i=0;i<count;i++){
      size=utf8_array[i]?sui_utf8_bytesize(utf8_array[i]):0;
      kvpb_serialize_internal(&p, &size, sizeof(size));
      if(utf8_array[i])
        memcpy(p,sui_utf8_get_text(utf8_array[i]),size);
      p=(char*)p+size;
    }
  }while(0);

  for(i=0;i<count;i++){
    if(utf8_array[i] != NULL)
      sui_utf8_dec_refcnt(utf8_array[i]);
  }

  return res?res:fullsize;
}

/**
 * kvpb_dinfos2dbuff - Serialize list of named dinfos values into the dynamic buffer
 * @pdbuf:	pointer to the dynamic buffer
 * @dilist:	pointer to the dinfo names to unique IDs conversion list
 * @options:    options for DINFO processing (%KVPB_DINFOS_IGNORE_ERR_VAL)
 *
 * The function takes list of named dinfos and their corresponding unique IDs
 * and stores values of dinfos into dynamic buffer.  The @dilist has already
 * be assigned by generated unique IDs - function kvpb_di_list_generate_ids()
 *
 * The function returns negative value if there is some problem
 */
int kvpb_dinfos2dbuff(ul_dbuff_t *pdbuf, kvpb_di_list_t *dilist, int options)
{
  kvpb_di_item_t *diitem;
  sui_dinfo_t *dinfo;
  int size, fullsize;
  kvpb_di_id_t diid;
  sui_typeid_t tinfo;
  int count;
  int startidx;
  int prev_len;
  void *p;
  int res;

  ul_list_for_each(kvpb_di_list_req, dilist, diitem){
    diid=diitem->id;
    dinfo=ns_getobj_dinfo_global(diitem->name.str);
    if(!dinfo){
      ul_logmsg("kvpb_dinfos2dbuff: no dinfo \"%s\"\n",diitem->name.str);
      continue;
    }

    count=dinfo->idxsize;
    if(!count)
      count=1;

    startidx=diitem->min_idx;
    if(startidx>=count){
      ul_logmsg("kvpb_dinfos2dbuff: dinfo \"%s\" index out of bounds\n",diitem->name.str);
      continue;
    }

    count-=startidx;

    if(startidx+count>diitem->max_idx)
      count=diitem->max_idx-startidx+1;

    tinfo=dinfo->tinfo;
    if(tinfo == SUI_TYPE_UTF8){
      res=kvpb_dinfos2dbuff_utf8(pdbuf, diid, dinfo, startidx, count, options);
      if(res>=0){
        ul_logdeb("kvpb_dinfos2dbuff: utf8 dinfo \"%s\" serialized to %d bytes\n",diitem->name.str,res);
      }else{
        ul_logerr("kvpb_dinfos2dbuff: utf8 dinfo \"%s\" store failed\n",diitem->name.str);
      }
      continue;
    }else{
      size=sui_typeid_size(sui_typeid_buffer_type(tinfo));
      if(size<=0){
        ul_logdeb("kvpb_dinfos2dbuff: dinfo \"%s\" size %d ?!?\n",diitem->name.str,size);
	continue;
      }
    }

    fullsize=sizeof(fullsize)+sizeof(tinfo)+sizeof(diid);
    if(startidx){
      diid|=KVPB_DI_ARRAYFL;
      fullsize+=sizeof(startidx);
    }
    fullsize+=size*count;

    ul_logdeb("kvpb_dinfos2dbuff: dinfo \"%s\" takes size %d\n",diitem->name.str,fullsize);

    prev_len=pdbuf->len;
    if(ul_dbuff_set_len(pdbuf, pdbuf->len+fullsize)<=0){
      return -1;
    }
    p=pdbuf->data + prev_len;

    kvpb_serialize_internal(&p, &fullsize, sizeof(fullsize));
    kvpb_serialize_internal(&p, &tinfo, sizeof(tinfo));
    kvpb_serialize_internal(&p, &diid, sizeof(diid));
    if(diid&KVPB_DI_ARRAYFL)
      kvpb_serialize_internal(&p, &startidx, sizeof(startidx));

    for(;count;count--,startidx++){
      if(!dinfo->rdval){
        ul_logerr("kvpb_dinfos2dbuff: dinfo \"%s\" read not allowed\n",diitem->name.str);
        return -1;
      }

      if(dinfo->rdval(dinfo, startidx, p)!=SUI_RET_OK){
        if(options & KVPB_DINFOS_IGNORE_ERR_VAL) {
	  memset(p, 0, size);
	}else {
          ul_logerr("kvpb_dinfos2dbuff: dinfo \"%s\" rdval idx %d error\n",diitem->name.str,startidx);
	  return -1;
	}
      }
      p=(char*)p+size;
    }
  }

  return 0;
}

/**
 * kvpb_dinfos2store - Save list of named dinfos values into the storage block
 * @store:	pointer to the keyval storage block description
 * @dilist:	pointer to the dinfo names to unique IDs conversion list
 * @keyid:	keyval ID of the value/block holding dinfos values
 *
 * The function takes list of named dinfos and their corresponding unique IDs
 * and stores values of dinfos into one keyval value item.
 *
 * The function returns negative value if there is some problem
 */
int kvpb_dinfos2store(struct kvpb_block *store, kvpb_di_list_t *dilist, int keyid)
{
  ul_dbuff_t dbuf;
  int res;

  ul_logdeb("kvpb_dinfos2store: storing to keyid 0x%x\n", keyid);
  kvpb_di_list_load_ids(dilist, store, KVPB_KEYID_DINFO_NAMES);
  if(store->flags&KVPB_DESC_RO)
    return -1;

  if(kvpb_di_list_generate_ids(dilist)<0){
    return -1;
  }

  if(dilist->update_names){
    ul_logdeb("kvpb_dinfos2store: updating names\n");
    if(kvpb_di_list_save_ids(dilist, store, KVPB_KEYID_DINFO_NAMES)<0)
      return -1;
  }

  ul_dbuff_init(&dbuf, 0);

  res=kvpb_dinfos2dbuff(&dbuf, dilist,0);

  if(res>=0){
    res=kvpb_set_key_dbuff(store, keyid, &dbuf);
    if(res<0){
      ul_logerr("kvpb_dinfos2store: kvpb_set_key returned %d\n",res);
    }
  }

  ul_dbuff_destroy(&dbuf);

  return res;
}

/**
 * kvpb_dbuff2dinfos - Load list of named dinfos values from the dynamic buffer
 * @pdbuf:	pointer to the dynamic buffer
 * @dilist:	pointer to the dinfo names to unique IDs conversion list
 * @options:    options for DINFO processing (%KVPB_DINFOS_IGNORE_ERR_VAL)
 *
 * The function takes list of named dinfos and their corresponding unique IDs
 * and loads dinfos values from dynamic buffer.
 *
 * The function returns negative value if there is some problem
 */
int kvpb_dbuff2dinfos(const ul_dbuff_t *pdbuf, kvpb_di_list_t *dilist, int options)
{
  kvpb_di_item_t *diitem;
  sui_dinfo_t *dinfo;
  const void *end, *p, *next;
  int size, fullsize;
  kvpb_di_id_t diid;
  sui_typeid_t tinfo;
  int startidx;
  int res=0;

  next=(uint8_t*)pdbuf->data;
  end=(uint8_t*)pdbuf->data+pdbuf->len;
  while(next<end-sizeof(fullsize)){
    p=next;
    kvpb_deserialize_internal(&p, &fullsize, sizeof(fullsize));
    if(!fullsize)
      break;
    next=(char*)next+fullsize;
    if(next>end)
      next=end;
    kvpb_deserialize_internal(&p, &tinfo, sizeof(tinfo));
    kvpb_deserialize_internal(&p, &diid, sizeof(diid));
    if(diid&KVPB_DI_ARRAYFL){
      kvpb_deserialize_internal(&p, &startidx, sizeof(startidx));
      diid&=~KVPB_DI_ARRAYFL;
    }else{
      startidx=0;
    }

    size=sui_typeid_size(sui_typeid_buffer_type(tinfo));

    diitem=kvpb_di_list_id_find(dilist,&diid);
    if(!diitem){
      res=-1;
      ul_loginf("kvpb_dbuff2dinfos: unknown ID (0x%x) in the store\n", diid);
      continue;
    }
    if(!(diitem->flags&KVPB_DI_RDWR))
      continue;

    dinfo=ns_getobj_dinfo_global(diitem->name.str);
    if(!dinfo){
      ul_loginf("kvpb_dbuff2dinfos: no dinfo \"%s\" in the namespace\n", diitem->name.str);
      res=-1;
      continue;
    }
    if(dinfo->tinfo!=tinfo){
      ul_loginf("kvpb_dbuff2dinfos: dinfo \"%s\" types mismatch\n", diitem->name.str);
      if (!(options & KVPB_DINFOS_IGNORE_DIFF_COMP_TYPES) ||
           (sui_typeid_buffer_type(tinfo)!=sui_typeid_buffer_type(dinfo->tinfo)) ||
           ((sui_typeid_buffer_type(tinfo)!=SUI_TYPE_LONG)&&
	    (sui_typeid_buffer_type(tinfo)!=SUI_TYPE_ULONG))) {
        ul_loginf("kvpb_dbuff2dinfos: types are incompatible or conversion is disallowed\n");
        res=-1;
        continue;
      }
    }

    if(!dinfo->wrval){
      res=-1;
      ul_logerr("kvpb_dbuff2dinfos: dinfo \"%s\" write not allowed\n",diitem->name.str);
      continue;
    }

    ul_logdeb("kvpb_dbuff2dinfos: dinfo \"%s\" retrieving\n",diitem->name.str);

    if(tinfo == SUI_TYPE_UTF8){
      for(;(char*)next-(char*)p>=sizeof(size); startidx++, p=(char*)p+size){
        utf8 *utf;
        kvpb_deserialize_internal(&p, &size, sizeof(size));
	if(size>(char*)next-(char*)p){
          ul_logerr("kvpb_dbuff2dinfos: utf8 dinfo \"%s\" size corrupted\n",diitem->name.str);
	  res=-1;
	  break;
	}

	if((startidx<diitem->min_idx) || (startidx>diitem->max_idx))
	  continue;

	if((dinfo->idxsize) && (dinfo->idxsize<=startidx))
          continue;

        utf=sui_utf8_dynamic(p, -1, size);
	if(!utf){
          ul_logerr("kvpb_dbuff2dinfos: utf8 dinfo \"%s\" idx %d error to create utf8 for %d bytes\n",
	             diitem->name.str,startidx,size);
	  res=-1;
          continue;
	}

	if(sui_wr_utf8(dinfo, startidx, &utf)!=SUI_RET_OK){
          if(!(options & KVPB_DINFOS_IGNORE_ERR_VAL)) {
            ul_logerr("kvpb_dbuff2dinfos: utf8 dinfo \"%s\" wrval idx %d error\n",diitem->name.str,startidx);
            res=-1;
	  }
	}

        sui_utf8_dec_refcnt(utf);
      }
    }else{
      int wrres = 0;
      for(;(char*)next-(char*)p>=size; startidx++,p=(char*)p+size){
	if((startidx<diitem->min_idx) || (startidx>diitem->max_idx))
          continue;

	if((dinfo->idxsize) && (dinfo->idxsize<=startidx))
          continue;
        if (tinfo==dinfo->tinfo) {
          wrres = dinfo->wrval(dinfo, startidx, p);
        } else {
          long tmpval = 0;
          wrres = sui_typeid_convert(dinfo->tinfo, &tmpval, tinfo, (void *)p, 1);
          if (wrres==SUI_RET_OK)
            wrres = dinfo->wrval(dinfo, startidx, &tmpval);
        }
	if(wrres!=SUI_RET_OK){
          if(!(options & KVPB_DINFOS_IGNORE_ERR_VAL)) {
            ul_logerr("kvpb_dbuff2dinfos: dinfo \"%s\" wrval idx %d error\n",diitem->name.str,startidx);
            res=-1;
          }
	}
      }
    }
  }

  return res;
}

/**
 * kvpb_store2dinfos - Load list of named dinfos values from the storage block
 * @store:	pointer to the keyval storage block description
 * @dilist:	pointer to the dinfo names to unique IDs conversion list
 * @keyid:	keyval ID of the value/block holding dinfos values
 *
 * The function takes list of named dinfos and their corresponding unique IDs
 * and loads dinfos values from keyval value item adressed by @keyid.
 *
 * The function returns negative value if there is some problem
 */
int kvpb_store2dinfos(struct kvpb_block *store, kvpb_di_list_t *dilist, int keyid)
{
  ul_dbuff_t dbuf;
  int res=0;

  ul_logdeb("kvpb_store2dinfos: loading from keyid 0x%x\n", keyid);
  if(kvpb_di_list_load_ids(dilist, store,KVPB_KEYID_DINFO_NAMES)<0){
    ul_logerr("kvpb_store2dinfos: kvpb_di_list_load_ids error\n");
    return -1;
  }

  ul_dbuff_init(&dbuf, 0);
  res=kvpb_get_key_dbuff(store, keyid, &dbuf);
  if((res<0) || !dbuf.len){
    ul_logerr("kvpb_store2dinfos: kvpb_get_key_dbuff returned %d \n", res);
    ul_dbuff_destroy(&dbuf);
    return res;
  }

  res=kvpb_dbuff2dinfos(&dbuf, dilist,0);

  ul_dbuff_destroy(&dbuf);
  return res;
}

/**
 * kvpb_compare_dbuff2dinfos - Compare list of named dinfos values against stored values
 * @pdbuf:	pointer to the dynamic buffer
 * @dilist:	pointer to the dinfo names to unique IDs conversion list
 *
 * The function takes list of named dinfos and their corresponding unique IDs
 * and compares dinfos values against values stored in the dynamic buffer.
 *
 * The function returns negative value if there is some problem
 * Positive value is returned if some difference is found, zero value indicates full match.
 */
int kvpb_compare_dbuff2dinfos(const ul_dbuff_t *pdbuf, kvpb_di_list_t *dilist)
{
  kvpb_di_item_t *diitem;
  sui_dinfo_t *dinfo;
  const void *end, *p, *next;
  int size, fullsize;
  kvpb_di_id_t diid;
  sui_typeid_t tinfo;
  ul_dbuff_t val_dbuf;
  int startidx;
  int res=0;
  int cmpres;

  ul_list_for_each(kvpb_di_list_req, dilist, diitem){
    if(!(diitem->flags&KVPB_DI_RDWR)) continue;
    diitem->flags|=KVPB_DI_DIFFER;
  }

  res=0;
  next=(uint8_t*)pdbuf->data;
  end=(uint8_t*)pdbuf->data+pdbuf->len;
  ul_dbuff_init(&val_dbuf, 0);
  while((next<end-sizeof(fullsize))&&!res){
    p=next;
    kvpb_deserialize_internal(&p, &fullsize, sizeof(fullsize));
    if(!fullsize)
      break;
    next=(char*)next+fullsize;
    if(next>end)
      next=end;
    kvpb_deserialize_internal(&p, &tinfo, sizeof(tinfo));
    kvpb_deserialize_internal(&p, &diid, sizeof(diid));
    if(diid&KVPB_DI_ARRAYFL){
      kvpb_deserialize_internal(&p, &startidx, sizeof(startidx));
      diid&=~KVPB_DI_ARRAYFL;
    }else{
      startidx=0;
    }
    size=sui_typeid_size(sui_typeid_buffer_type(tinfo));

    diitem=kvpb_di_list_id_find(dilist,&diid);
    if(!diitem){
      res=-1;
      continue;
    }
    if(!(diitem->flags&KVPB_DI_RDWR)){
      if(!res) res=1;
      continue;
    }

    dinfo=ns_getobj_dinfo_global(diitem->name.str);
    if(!dinfo){
      res=-1;
      continue;
    }
    if(dinfo->tinfo!=tinfo){
      res=-1;
      continue;
    }

    ul_logdeb("kvpb_compare_store2dinfos: dinfo \"%s\" retrieving\n",diitem->name.str);

    if(startidx<diitem->min_idx){
      if(!res) res=1;
      continue;
    }

    if(val_dbuf.capacity<=size){
      val_dbuf.len=size;
    }else{
      if(ul_dbuff_prep(&val_dbuf, size)!=size){
	res=-1;
	continue;
      }
    }

    cmpres=0;
    if(tinfo == SUI_TYPE_UTF8){
      for(;(char*)next-(char*)p>=sizeof(size); startidx++, p=(char*)p+size){
        utf8 *utf = NULL;
        kvpb_deserialize_internal(&p, &size, sizeof(size));
	if(size>(char*)next-(char*)p){
          ul_logerr("kvpb_store2dinfos: utf8 dinfo \"%s\" size corrupted\n",diitem->name.str);
	  res=-1;
	  break;
	}

	if((startidx<diitem->min_idx) || (startidx>diitem->max_idx)){
          cmpres=1;
          break;
	}

	if((dinfo->idxsize) && (dinfo->idxsize<=startidx)){
          cmpres=1;
          break;
	}

	if(sui_rd_utf8(dinfo, startidx, &utf)!=SUI_RET_OK){
          ul_logerr("kvpb_store2dinfos: dinfo \"%s\" wrval idx %d error\n",diitem->name.str,startidx);
          cmpres=1;
          res=-1;
	  break;
	}

        if(size!=sui_utf8_bytesize(utf)){
          ul_logdeb("kvpb_compare_store2dinfos: utf8 sizes differ %d != %d\n",size,sui_utf8_bytesize(utf));
          cmpres=1;
          break;
	}

        if(memcmp(sui_utf8_get_text(utf),p,size)){
          ul_logdeb("kvpb_compare_store2dinfos: utf8 content differ\n");
          cmpres=1;
          break;
	}

        sui_utf8_dec_refcnt(utf);
      }
    }else{
      for(;(char*)next-(char*)p>=size; startidx++,p=(char*)p+size){
	if(startidx>diitem->max_idx){
          cmpres=1;
          break;
	}

	if((dinfo->idxsize) && (dinfo->idxsize<=startidx)){
          cmpres=1;
          break;
	}

	if(dinfo->rdval(dinfo, startidx, val_dbuf.data)!=SUI_RET_OK){
          ul_logerr("kvpb_compare_store2dinfos: dinfo \"%s\" rdval idx %d error\n",diitem->name.str,startidx);
	  cmpres=1;
          res=-1;
	  break;
	}

	cmpres=memcmp(p,val_dbuf.data,size);
	if(cmpres) break;
      }
    }
    if(cmpres){
      if(!res) res=1;
    }else{
      diitem->flags&=~KVPB_DI_DIFFER;
    }
  }

  if(!res){
    ul_list_for_each(kvpb_di_list_req, dilist, diitem){
      if(diitem->flags&KVPB_DI_DIFFER)
	res=1;
    }
  }

  ul_dbuff_destroy(&val_dbuf);
  return res;
}

/**
 * kvpb_compare_store2dinfos - Compare list of named dinfos values against stored values
 * @store:	pointer to the keyval storage block description
 * @dilist:	pointer to the dinfo names to unique IDs conversion list
 * @keyid:	keyval ID of the value/block holding dinfos values
 *
 * The function takes list of named dinfos and their corresponding unique IDs
 * and compares dinfos values against values stored in the keyval value item
 * adressed by @keyid.
 *
 * The function returns negative value if there is some problem
 * Positive value is returned if some difference is found, zero value indicates full match.
 */
int kvpb_compare_store2dinfos(struct kvpb_block *store, kvpb_di_list_t *dilist, int keyid)
{
  ul_dbuff_t dbuf;
  int res=0;

  ul_logdeb("kvpb_compare_store2dinfos: comparing with keyid 0x%x\n", keyid);
  if(kvpb_di_list_load_ids(dilist, store,KVPB_KEYID_DINFO_NAMES)<0){
    ul_logerr("kvpb_compare_store2dinfos: kvpb_di_list_load_ids error\n");
    return -1;
  }

  ul_dbuff_init(&dbuf, 0);
  res=kvpb_get_key_dbuff(store, keyid, &dbuf);
  if((res<0) || !dbuf.len){
    ul_logerr("kvpb_compare_store2dinfos: kvpb_get_key_dbuff returned %d \n", res);
    ul_dbuff_destroy(&dbuf);
    if(res==0) {
      /*
       * ensure, that difference is reported if the store holds empty block
       * and there is some DINFO to conditionally store
       */
      kvpb_di_item_t *diitem;
      ul_list_for_each(kvpb_di_list_req, dilist, diitem)
        if(diitem->flags&KVPB_DI_RDWR) return 1;
    }
    return res;
  }

  res=kvpb_compare_dbuff2dinfos(&dbuf, dilist);

  ul_dbuff_destroy(&dbuf);
  ul_logdeb("kvpb_compare_store2dinfos: compare keyid 0x%x result %d\n", keyid, res);
  return res;
}
