#include <string.h>
#include <system_def.h>
#include <bbconf_info.h>
#include <keyvalpb.h>
#include <stdio.h>
#include <stdlib.h>
#include <utils.h>
#include <flashlib.h>
#include <malloc.h>
#include <keyval_cmd.h>
#include "keyval_flash.h"

#if 1

#include <ul_log.h>

UL_LOG_CUST(ulogd_keyval_flash)

#define kvpb_logerr ul_logerr
#define kvpb_logmsg ul_logmsg
#define kvpb_logdeb ul_logdeb

#else

#include <stdarg.h>

int kvpb_debug=1; 

static inline
void kvpb_logerr(const char *format, ...)
{
  va_list ap;
  va_start (ap, format);
  vprintf(format, ap);
  va_end (ap);
}

static inline
void kvpb_logmsg(const char *format, ...)
{
  va_list ap;
  va_start (ap, format);
  vprintf(format, ap);
  va_end (ap);
}

static inline
void kvpb_logdeb(const char *format, ...)
{
  va_list ap;
  if(!kvpb_debug)
    return;
  va_start (ap, format);
  vprintf(format, ap);
  va_end (ap);
}
#endif

flash_alg_info_t kvpb_flash_alg_data;

flash_alg_info_t *kvpb_flash_alg=&kvpb_flash_alg_data;

kvpb_block_t kvpb_flash_store_data;

kvpb_block_t *kvpb_flash_store=&kvpb_flash_store_data;

#define KVPB_FLASH_BUF_SIZE 0x1000

uint8_t kvpb_flash_buf[KVPB_FLASH_BUF_SIZE];

/**
 * kvpb_flash_erase - FLASH memory specific function to erase KVPB region
 * @store: Pointer to the KVPB access information/state structure
 * @base: Base address of erased region inside parameter block data region
 * @size: Number of bytes to erase
 *
 * This function should not be called directly. It provides FLASH specific
 * implementation for kvpb_block_erase() wrapper function.
 *
 * Return Value: Negative value indicates operation fault.
 * File: keyval_flash.c
 */
int kvpb_flash_erase(struct kvpb_block *store, void *base,int size)
{
  int res;
  if(store->flags&KVPB_DESC_RO)
    return -1;

  if(((unsigned long)base & 3)||(size&3)){
    kvpb_logerr("ALIGNERROR: kvpb_flash_erase(%p, %d)\n", base, size);
  } else {
    kvpb_logdeb("kvpb_flash_erase(%p, %d)\n", base, size);
  }
  if(ulogd_keyval_flash.level>=UL_LOGL_TRASH)
    sleep(1);

  res=strataflash_erase_x16(kvpb_flash_alg, (flash_addr_t)base, size);
  if(res<0){
    store->flags|=KVPB_DESC_RO;
    kvpb_logerr("ERROR: strataflash_erase_x16 returned %d\n",res);
  }
  if(ulogd_keyval_flash.level>=UL_LOGL_TRASH)
    sleep(1);
  kvpb_logdeb("kvpb_flash_erase result %d\n", res);
  return res;
}

/**
 * kvpb_flash_copy - FLASH memory specific function to copy data into KVPB storage
 * @store: Pointer to the KVPB access information/state structure
 * @dst: Address of data destination pointing inside mapped parameter block data region
 * @src: Address of data source pointing inside mapped parameter block data or RAM memory
 * @len: Number of bytes to transfer
 *
 * This function should not be called directly. It provides FLASH specific
 * implementation for kvpb_block_copy() wrapper function.
 *
 * Return Value: Negative value indicates operation fault.
 * File: keyval_flash.c
 */
int kvpb_flash_copy(struct kvpb_block *store, void *dst,const void *src,int len)
{
  int res;

  if(store->flags&KVPB_DESC_RO)
    return -1;

  if(((unsigned long)dst & 3)||((unsigned long)src & 3)||(len&3)){
    kvpb_logerr("ALIGNERROR: kvpb_flash_copy(%p, %p, %d)\n", dst, src, len);;
  } else {
    kvpb_logdeb("kvpb_flash_copy(%p, %p, %d)\n", dst, src, len);
  }
  if(ulogd_keyval_flash.level>=UL_LOGL_TRASH)
    sleep(1);

  if(!len)
    return 0;

  if(((long)src^(long)dst)&~0xffffff){
    res=strataflash_prog_x16(kvpb_flash_alg, (flash_addr_t)dst, src, len);
  }else{
    int blen;
    int alignoffs;
    blen=KVPB_FLASH_BUF_SIZE-((unsigned long)dst&(KVPB_FLASH_BUF_SIZE-1));
    alignoffs=(unsigned long)dst&3;
    do{
      if(blen>len)
        blen=len;
      memcpy(kvpb_flash_buf+alignoffs, src, blen);
      res=strataflash_prog_x16(kvpb_flash_alg, (flash_addr_t)dst, kvpb_flash_buf+alignoffs, blen);
      dst=(uint8_t*)dst+blen;
      src=(uint8_t*)src+blen;
      len-=blen;
      if(res<0)
        break;
      blen=KVPB_FLASH_BUF_SIZE;
      alignoffs=0;
    } while(len);
  }
  if(res<0){
    store->flags|=KVPB_DESC_RO;
    kvpb_logerr("ERROR: strataflash_prog_x16 returned %d\n",res);
  }
  if(ulogd_keyval_flash.level>=UL_LOGL_TRASH)
    sleep(1);
  kvpb_logdeb("kvpb_flash_copy result %d\n", res);
  return res;
}

kvpb_block_t kvpb_flash_store_data={
  .base=(void*)KVPB_FLASH_AREAS_START,
  .size=KVPB_FLASH_AREA_SIZE,
  .flags=KVPB_DESC_DOUBLE|KVPB_DESC_ALIGN4,
 #ifndef  KVPB_MINIMALIZED
  .erase=kvpb_flash_erase,
  .copy=kvpb_flash_copy,
 #endif  /* KVPB_MINIMALIZED */
};

/**
 * kvpb_mem_erase - Generic RAM memory function to erase KVPB region
 * @store: Pointer to the KVPB access information/state structure
 * @base: Base address of erased region inside parameter block data region
 * @size: Number of bytes to erase
 *
 * This function should not be called directly. It provides generic RAM memory
 * implementation for kvpb_block_erase() wrapper function.
 *
 * Return Value: Negative value indicates operation fault.
 * File: keyval_flash.c
 */
int kvpb_mem_erase(struct kvpb_block *store, void *base,int size)
{
  if(store->flags&KVPB_DESC_RO)
    return -1;

  if(((unsigned long)base & 3)||(size&3)){
    kvpb_logerr("ALIGNERROR: kvpb_mem_errase(%p, %d)\n", base, size);
  } else {
    kvpb_logdeb("kvpb_mem_errase(%p, %d)\n", base, size);
  }
  memset(base,0xff,size);
  
  return 0;
}

/**
 * kvpb_mem_copy - Generic RAM function to copy data into KVPB storage
 * @store: Pointer to the KVPB access information/state structure
 * @dst: Address of data destination pointing inside mapped parameter block data region
 * @src: Address of data source pointing inside mapped parameter block data or RAM memory
 * @len: Number of bytes to transfer
 *
 * This function should not be called directly. It provides generic RAM memory
 * implementation for kvpb_block_copy() wrapper function.
 *
 * Return Value: Negative value indicates operation fault.
 * File: keyval_flash.c
 */
int kvpb_mem_copy(struct kvpb_block *store, void *dst,const void *src,int len)
{
  if(store->flags&KVPB_DESC_RO)
    return -1;

  if(((unsigned long)dst & 3)||((unsigned long)src & 3)||(len&3)){
    kvpb_logerr("ALIGNERROR: kvpb_mem_copy(%p, %p, %d)\n", dst, src, len);;
  } else {
    kvpb_logdeb("kvpb_mem_copy(%p, %p, %d)\n", dst, src, len);
  }
  memcpy(dst,src,len);
  return 0;
}

kvpb_block_t kvpb_mem_store_data={
  .base=NULL,
  .size=0x400,
  .flags=KVPB_DESC_DOUBLE|KVPB_DESC_ALIGN4,
 #ifndef  KVPB_MINIMALIZED
  .erase=kvpb_mem_erase,
  .copy=kvpb_mem_copy,
 #endif  /* KVPB_MINIMALIZED */
};

/**
 * kvpb_get_forshell - Shell function to retrieve value for given key
 * @argc: main() style argument count
 * @argv: main() style arguments values
 *
 * Return Value: Non-zero value inform about unsuccessful operation.
 * File: keyval_flash.c
 */
int kvpb_get_forshell(int argc, char **argv)
{
  int res;
  char *line;
  if(concat_args2line(&line,argc,argv)<0){
    fprintf(stderr, "kvpb_get_forshell: concat_args2line failed\n");
    return 3;
  }
  res=kvpb_get_cmd(kvpb_flash_store, line);
  free(line);
  if(res<0){
    printf("ERROR: kvpb_get_cmd returned %d\n", res);
  }
  return res;
}

/**
 * kvpb_set_forshell - Shell function to set value for given key
 * @argc: main() style argument count
 * @argv: main() style arguments values
 *
 * Return Value: Non-zero value inform about unsuccessful operation.
 * File: keyval_flash.c
 */
int kvpb_set_forshell(int argc, char **argv)
{
  int res;
  char *line;
  if(concat_args2line(&line,argc,argv)<0){
    fprintf(stderr, "kvpb_set_forshell: concat_args2line failed\n");
    return 3;
  }
  res=kvpb_set_cmd(kvpb_flash_store, line);
  free(line);
  if(res<0){
    printf("ERROR: kvpb_set_cmd returned %d\n", res);
  }
  return res<0?1:0;
}

/**
 * kvpb_del_forshell - Shell function to delete given key and values
 * @argc: main() style argument count
 * @argv: main() style arguments values
 *
 * Return Value: Non-zero value inform about unsuccessful operation.
 * File: keyval_flash.c
 */
int kvpb_del_forshell(int argc, char **argv)
{
  int res;
  char *line;
  if(concat_args2line(&line,argc,argv)<0){
    fprintf(stderr, "kvpb_del_forshell: concat_args2line failed\n");
    return 3;
  }
  res=kvpb_del_cmd(kvpb_flash_store, line);
  free(line);
  if(res<0){
    printf("ERROR: kvpb_del_cmd returned %d\n", res);
  }
  return res<0?1:0;
}

/**
 * kvpb_list_forshell - Shell function to list KVPB storage content
 * @argc: main() style argument count
 * @argv: main() style arguments values
 *
 * Return Value: Non-zero value inform about unsuccessful operation.
 * File: keyval_flash.c
 */
int kvpb_list_forshell(int argc, char **argv)
{
  int res;

  printf("Store list:\n");
  res=kvpb_store_list(kvpb_flash_store, 0);
  printf("--- end ---\n");
  if(res<0){
    return -res;
  }

  res=kvpb_check(kvpb_flash_store,0);
  if(res!=0){
    printf("ERROR: kvpb_check returned %d\n", res);
    return res<0?2:1;
  }
    
  return 0;
}

/**
 * kvpb_stress_forshell - Shell function to stress test KVPB storage
 * @argc: main() style argument count
 * @argv: main() style arguments values
 *
 * Return Value: Non-zero value inform about unsuccessful operation.
 * File: keyval_flash.c
 */
int kvpb_stress_forshell(int argc, char **argv)
{
  int cycles = 1000;
  int res;

  if(argc >= 2)
    cycles = atoi(argv[1]);

  res = kvpb_stress_test(kvpb_flash_store, cycles, 10, 300, 1000, 1020);

  return res>=0?0:-res;
}

/**
 * kvpb_flash_store_init - Initialize default KVPB FLASH storage
 *
 * Return Value: Non-zero value inform about problems to fulfill operation.
 *	Negative value means fatal problem and storage cannot be accessed.
 * File: keyval_flash.c
 */
int kvpb_flash_store_init(void)
{
  unsigned long val;
  int res;

  if(bbconf_get_param(BBCONF_PTTAG_BBVER, &val)<0) {
    kvpb_logerr("cannot get BBCONF parameters\n");
  } else {
    if(bbconf_get_param(BBCONF_PTTAG_KVPB_SIZE, &val)>0)
      kvpb_flash_store_data.size = val;
    res=bbconf_get_param(BBCONF_PTTAG_KVPB_BYCFI, &val);
    if((res>0) && (val==1)) {
      /*flash_alg_default->getinfo*/
      res = strataflash_getinfo_x16(kvpb_flash_alg, BBCONF_FLASH_START,
                                   FLASH_INFO_CFI_SIZE_LOG2, &val);
      if(res<0) {
        kvpb_logerr("kvpb get CFI flash info failed\n");
	return -1;
      } else {
        val=BBCONF_FLASH_START+(1<<val)-2*kvpb_flash_store_data.size;
        kvpb_flash_store_data.base=(uint8_t *)val;
	if((unsigned long)kvpb_flash_store_data.base < BBCONF_FLASH_START) {
          kvpb_logerr("incorrect KVPB block start 0x%08lx\n", val);
	  return -1;
	}
      }
    } else {
      if(bbconf_get_param(BBCONF_PTTAG_KVPB_START, &val)>0)
        kvpb_flash_store_data.base=(uint8_t *)val;
    }
  }
  kvpb_logmsg("KVPB block starts at 0x%08lx size 0x%08lx\n",
              (long)kvpb_flash_store_data.base, (long)kvpb_flash_store_data.size);

  if(1){
    kvpb_flash_store=&kvpb_flash_store_data;
    return kvpb_store_init(kvpb_flash_store, kvpb_flash_store_data.base,
  			   kvpb_flash_store_data.size, KVPB_DESC_DOUBLE|KVPB_DESC_ALIGN4, 1);
  }else{
    kvpb_flash_store=&kvpb_mem_store_data;
    return kvpb_store_init(kvpb_flash_store, malloc(2*0x400),
  			   0x400, KVPB_DESC_DOUBLE|KVPB_DESC_ALIGN4, 1);
  }
}

/**
 * kvpb_flash_block - Get pointer to the FLASH KVPB access structure
 * @name: Optional block ASCII name or %NULL
 * @number: Sequential number/minor of the KVPB storage
 *
 * Return Value: Pointer to the KVPB access information/state structure
 *	according to provided parameters or %NULL if corresponding block does
 *	not exist.
 * File: keyval_flash.c
 */
struct kvpb_block *kvpb_flash_block(char *name, int number)
{
  return kvpb_flash_store;
}
