/*******************************************************************
  Key Value Persistent Storage

  keyval_passwd.h	- password and capabilities management

  (C) Copyright 2003-2005 by Pavel Pisa - Originator

  The uLan utilities library can be used, copied and modified under
  next licenses
    - GPL - GNU General Public License
    - LGPL - GNU Lesser General Public License
    - MPL - Mozilla Public License
    - and other licenses added by project originators
  Code can be modified and re-distributed under any combination
  of the above listed licenses. If contributor does not agree with
  some of the licenses, he/she can delete appropriate line.
  Warning, if you delete all lines, you are not allowed to
  distribute source code and/or binaries utilizing code.
  
  See files COPYING and README for details.

 *******************************************************************/

#include <inttypes.h>
#include <string.h>
#include <keyvalpb.h>
#include <stdio.h>
#include <malloc.h>
#include <keyval_dilist.h>
#include <keyval_dbuff.h>
#include <keyval_flash.h>
#include <keyval_id.h>

#include <keyval_passwd.h>
#include <sha1.h>

#define KEYVAL_PASSWD_LEN 20

typedef struct {
  uint8_t bytes[KEYVAL_PASSWD_LEN];
} keyval_passwd_hash_t;

int keyval_passwd_initialized=0;

static ul_dbuff_t keyval_passwd_dbuff;

static ul_dbuff_t keyval_caps_dbuff;

static kvpb_di_list_t keyval_passwd_names;

static kvpb_block_t *keyval_passwd_store;

/**
 * keyval_passwd_hash_get - Retrieve hash of password corresponding to given UID
 * @uid: User ID
 * @hash: Pointer to the buffer to store password hash to
 *
 * Return Value: Zero value indicates successful operation
 *	or -1 if no hash can be retrieved for given UID.
 * File: keyval_passwd.c
 */
int keyval_passwd_hash_get(int uid, keyval_passwd_hash_t* hash)
{
  int pos=uid*sizeof(keyval_passwd_hash_t);
  if(keyval_passwd_dbuff.len<pos+sizeof(keyval_passwd_hash_t))
    return -1;
  memcpy(hash,keyval_passwd_dbuff.data+pos,sizeof(keyval_passwd_hash_t));
  return 0;
}

/**
 * keyval_passwd_hash_set - Set hash of the new password for given UID
 * @uid: User ID
 * @hash: Pointer to the buffer holding new hash value
 *
 * Return Value: Zero value indicates successful operation
 *	or -1 if hash cannot be set for given UID.
 * File: keyval_passwd.c
 */
int keyval_passwd_hash_set(int uid, const keyval_passwd_hash_t* hash)
{
  int pos=uid*sizeof(keyval_passwd_hash_t);
  int len=pos+sizeof(keyval_passwd_hash_t);
  int res;
  
  if(keyval_passwd_dbuff.len<len){
    int old_len=keyval_passwd_dbuff.len;
    if(ul_dbuff_set_len(&keyval_passwd_dbuff,len)!=len)
      return -1;
    memset(keyval_passwd_dbuff.data+old_len,0x55,len-old_len);
  }
  memcpy(keyval_passwd_dbuff.data+pos,hash,sizeof(keyval_passwd_hash_t));
  res=kvpb_set_key_dbuff(keyval_passwd_store, KVPB_KEYID_SECURITY_PASSWORDS, &keyval_passwd_dbuff);
  return (res<0)?-1:0;
}

/**
 * keyval_uid_get - Get UID for given user name
 * @name_str: Pointer to the ASCI character array holding user name
 * @name_len: Length of the user name
 *
 * Return Value: User identification (UID) or -1 if user name has not been found in the list.
 * File: keyval_passwd.c
 */
int keyval_uid_get(char *name_str, int name_len)
{
  kvpb_di_id_t uid;
  if(kvpb_di_list_name2id(&keyval_passwd_names, name_str, name_len, &uid)<0)
    return -1;
  return uid;
}

/**
 * keyval_uid_new - Allocate and register free UID for given user name
 * @name_str: Pointer to the ASCI character array holding user name
 * @name_len: Length of the user name
 *
 * Return Value: User identification (UID) or -1 if user has been already registered.
 * File: keyval_passwd.c
 */
int keyval_uid_new(char *name_str, int name_len)
{
  int res;

  if(keyval_uid_get(name_str, name_len)>=0)
    return -1;

  res=kvpb_di_list_add(&keyval_passwd_names, name_str, name_len, 0, 0, 0);
  if(res<0)
    return res;

  res=kvpb_di_list_generate_ids(&keyval_passwd_names);
  if(res<0)
    return res;
  
  res=kvpb_di_list_save_ids(&keyval_passwd_names, keyval_passwd_store, KVPB_KEYID_SECURITY_PASSNAMES);
  if(res<0)
    return res;

  return keyval_uid_get(name_str, name_len);
}


/**
 * keyval_caps_get - Retrieve set of the capabilities corresponding to given UID
 * @uid: User ID
 * @caps: Pointer to the buffer to store capabilities to
 *
 * Return Value: Zero value indicates successful operation
 *	or -1 if capabilities cannot be retrieved for given UID.
 * File: keyval_passwd.c
 */
int keyval_caps_get(int uid, keyval_caps_t* caps)
{
  int pos=uid*sizeof(keyval_caps_t);
  *caps=0;
  if(keyval_caps_dbuff.len<pos+sizeof(keyval_caps_t))
    return -1;
  memcpy(caps,keyval_caps_dbuff.data+pos,sizeof(keyval_caps_t));
  return 0;
}

/**
 * keyval_caps_set - Set capabilities for given UID
 * @uid: User ID
 * @caps: Pointer to the buffer holding new capabilities
 *
 * Return Value: Zero value indicates successful operation
 *	or -1 if capabilities cannot be set for given UID.
 * File: keyval_passwd.c
 */
int keyval_caps_set(int uid, const keyval_caps_t* caps)
{
  int pos=uid*sizeof(keyval_caps_t);
  int len=pos+sizeof(keyval_caps_t);
  int res;
  
  if(keyval_caps_dbuff.len<len){
    int old_len=keyval_caps_dbuff.len;
    if(ul_dbuff_set_len(&keyval_caps_dbuff,len)!=len)
      return -1;
    memset(keyval_caps_dbuff.data+old_len,0x55,len-old_len);
  }
  memcpy(keyval_caps_dbuff.data+pos,caps,sizeof(keyval_caps_t));
  res=kvpb_set_key_dbuff(keyval_passwd_store, KVPB_KEYID_SECURITY_PASSCAPS, &keyval_caps_dbuff);
  return (res<0)?-1:0;
}

/**
 * keyval_uid_name_get - Get user name for given UID
 * @uid: User ID
 * @pname: Pointer to the place where pointer to the newly allocated buffer
 *	holding retrieved name should be stored
 * @pname_len: Points to place where to store user name length
 *
 * Return Value: Zero value indicates successful operation
 *	or -1 if name for given UID cannot be found or there is not enough memory.
 * File: keyval_passwd.c
 */
int keyval_uid_name_get(int uid, char **pname, int *pname_len)
{
  kvpb_di_id_t di_uid=uid;
  kvpb_di_item_t *di_item;
  char *name;

  *pname=NULL;

  di_item=kvpb_di_list_id_find(&keyval_passwd_names, &di_uid);
  if(di_item==NULL)
    return -1;

  if(pname){
    name=malloc(di_item->name.len+1);
    if(!name)
      return -1;
    name[di_item->name.len]=0;
    memcpy(name,di_item->name.str,di_item->name.len);
    *pname=name;
  }

  if(*pname_len)
    *pname_len=di_item->name.len;

  return 0;
}

typedef struct {
  int(*callfnc)(void *context, int uid, const char *name, int name_len);
  void *context;
} keyval_uid_for_each_helper_t;

static int keyval_uid_for_each_callfnc(void *context, kvpb_di_item_t *diitem)
{
  keyval_uid_for_each_helper_t *phelper=(keyval_uid_for_each_helper_t *)context;
  return (*phelper->callfnc)(phelper->context, diitem->id, diitem->name.str, diitem->name.len);
}

/**
 * keyval_uid_for_each - Call specified function for all registered user names
 * @callfnc: Function which should be called for all users
 * @context: Function specific passed to the @callfnc function
 *
 * Return Value: Negative value indicates premature iteration break caused by negative
 *	value returned by @callfnc function.
 * File: keyval_passwd.c
 */
int keyval_uid_for_each(int(*callfnc)(void *context, int uid, const char *name, int name_len), void *context)
{
  int res;
  keyval_uid_for_each_helper_t helper;

  helper.context=context;
  helper.callfnc=callfnc;

  res=kvpb_di_list_for_each(&keyval_passwd_names, &helper, keyval_uid_for_each_callfnc);

  return res;
}

/**
 * keyval_passwd_init - Initialize subsystem and retrieve stored database
 * @name: Optional ASCII name of KVPB block holding password database or %NULL
 * @number: Sequential number/minor of the KVPB storage
 *
 * Return Value: Negative value indicates fault in the user database initialization.
 * File: keyval_passwd.c
 */
int keyval_passwd_init(char *name, int number)
{
  int res, ret;

  if(keyval_passwd_initialized)
    return 0;

  ul_dbuff_init(&keyval_passwd_dbuff, 0);
  ul_dbuff_init(&keyval_caps_dbuff, 0);

  kvpb_di_list_init(&keyval_passwd_names);

  kvpb_di_list_enforce_id(&keyval_passwd_names, "producer",  -1, KVSEC_UID_PRODUCER);
  kvpb_di_list_enforce_id(&keyval_passwd_names, "service",   -1, KVSEC_UID_SERVICE);
  kvpb_di_list_enforce_id(&keyval_passwd_names, "anonymous", -1, KVSEC_UID_ANONYMOUS);
  
  keyval_passwd_store=kvpb_flash_block(name, number);
  if(!keyval_passwd_store)
    return -1;
  
  ret=kvpb_di_list_load_ids(&keyval_passwd_names, keyval_passwd_store, KVPB_KEYID_SECURITY_PASSNAMES);

  res=kvpb_get_key_dbuff(keyval_passwd_store, KVPB_KEYID_SECURITY_PASSWORDS, &keyval_passwd_dbuff);
  if(res<ret)
    ret=res;
  
  res=kvpb_get_key_dbuff(keyval_passwd_store, KVPB_KEYID_SECURITY_PASSCAPS, &keyval_caps_dbuff);
  if(res<ret)
    ret=res;

  keyval_passwd_initialized=1;
  
  return ret;
}

/**
 * keyval_passwd_make_hash - Make has for given ASCII password
 * @hash: Pointer to buffer prepared to receive computed hash
 * @passwd: Pointer to character array holding password
 * @passwd_len: Length of the password
 * @single: Password is already hashed once.
 *
 * Return Value: Negative value indicates fault in hash computation.
 * File: keyval_passwd.c
 */
int keyval_passwd_make_hash(keyval_passwd_hash_t* hash, char *passwd, int passwd_len, int single)
{
  SHA_CTX *ctx;

  if(passwd_len<=0)
    passwd_len=strlen(passwd);

  ctx=malloc(sizeof(SHA_CTX));
  if(!ctx) return -1;

  do{
    SHA1_Init(ctx);
    SHA1_Update(ctx, passwd, passwd_len);
    SHA1_Final(hash->bytes, ctx);
    if(single)
      break;
    passwd=(char*)hash->bytes;
    passwd_len=KEYVAL_PASSWD_LEN;
    single=1;
  }while(1);

  free(ctx);
  
  return 0;
}

/**
 * keyval_passwd_check - Check user provided password
 * @uid: User ID
 * @passwd: Pointer to character array holding password
 * @passwd_len: Length of the password
 * @single: Password is already hashed once.
 *
 * Return Value: Negative value indicates incorrect password or fault.
 *	-1 .. password does not match, -2 .. cannot retrieve hash for given UID,
 *	-3 .. hash computation error.
 * File: keyval_passwd.c
 */
int keyval_passwd_check(int uid, char *passwd, int passwd_len, int single)
{
  keyval_passwd_hash_t hash1;
  keyval_passwd_hash_t hash2;
  
  if(keyval_passwd_make_hash(&hash1, passwd, passwd_len, single)<0)
    return -3;
  
  if(keyval_passwd_hash_get(uid, &hash2)<0)
    return -2;
  
  if(memcmp(hash1.bytes,hash2.bytes,KEYVAL_PASSWD_LEN))
    return -1;
    
  return 0;

}

/**
 * keyval_passwd_set - Set new password for given user
 * @uid: User ID
 * @passwd: Pointer to character array holding password
 * @passwd_len: Length of the password
 * @single: Password is already hashed once.
 *
 * Return Value: Negative value indicates fault in password set operation.
 *	-1 .. password cannot be stored, -3 .. hash computation error.
 * File: keyval_passwd.c
 */
int keyval_passwd_set(int uid, char *passwd, int passwd_len, int single)
{
  keyval_passwd_hash_t hash;
  
  if(keyval_passwd_make_hash(&hash, passwd, passwd_len, single)<0)
    return -3;
  
  if(keyval_passwd_hash_set(uid, &hash)<0)
    return -1;
  
  return 0;
}

/**
 * keyval_check_password_forshell - Shell function to check password for given user
 * @argc: main() style argument count
 * @argv: main() style arguments values
 *
 * Return Value: Non-zero value inform about unsuccessful operation.
 * File: keyval_passwd.c
 */
int keyval_check_password_forshell(int argc, char **argv)
{
  int uid;
  int res;

  if(!keyval_passwd_initialized)
    keyval_passwd_init("", 0);

  if(argc<3){
    printf("usage: %s username password\n",argv[0]);
    return 2;
  }
  
  uid=keyval_uid_get(argv[1], -1);
  if(uid<0){
    printf("%s: no user %s found\n",argv[0],argv[1]);
    return 2;
  }

  res=keyval_passwd_check(uid, argv[2], -1, 0);
  if(res<0){
    printf("%s: keyval_passwd_check returned %d\n",argv[0],res);
    return 1;
  }

  return 0;
}

/**
 * keyval_add_user_forshell - Shell function to add new user
 * @argc: main() style argument count
 * @argv: main() style arguments values
 *
 * Return Value: Non-zero value inform about unsuccessful operation.
 * File: keyval_passwd.c
 */
int keyval_add_user_forshell(int argc, char **argv)
{
  int uid;

  if(!keyval_passwd_initialized)
    keyval_passwd_init("", 0);

  if(argc<2){
    printf("usage: %s username\n",argv[0]);
    return 2;
  }
  
  uid=keyval_uid_new(argv[1], -1);
  if(uid<0){
    printf("%s: cannot create user %s\n",argv[0],argv[1]);
    return 2;
  }
  
  return 0;
}

/**
 * keyval_set_password_forshell - Shell function to set new password for given user
 * @argc: main() style argument count
 * @argv: main() style arguments values
 *
 * Return Value: Non-zero value inform about unsuccessful operation.
 * File: keyval_passwd.c
 */
int keyval_set_password_forshell(int argc, char **argv)
{
  int uid;
  int res;

  if(!keyval_passwd_initialized)
    keyval_passwd_init("", 0);

  if(argc<3){
    printf("usage: %s username password\n",argv[0]);
    return 2;
  }
  
  uid=keyval_uid_get(argv[1], -1);
  if(uid<0){
    printf("%s: no user %s found\n",argv[0],argv[1]);
    return 2;
  }

  res=keyval_passwd_set(uid, argv[2], -1, 0);
  if(res<0){
    printf("%s: keyval_passwd_set returned %d\n",argv[0],res);
    return 1;
  }

  return 0;
}

